/**
 * 
  xd87.ino , Main sketch of "Hobo-nicola xd87 keyboard" using Hobo-nicola keyboard and adapter library.
  Copyright (c) 2020 Takeshi Higasa, okiraku-camera.tokyo
  
  This file is part of "Hobo-nicola keyboard and adapter".

  "Hobo-nicola keyboard and adapter" is free software: you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation, either version 3 of the License, or
  (at your option) any later version.

  "Hobo-nicola keyboard and adapter" is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with "Hobo-nicola keyboard and adapter".  If not, see <http://www.gnu.org/licenses/>.

    Included in hoboNicola 1.5.2.		Sep. 20. 2021.
		            hoboNicola 1.6.1		Feb.22.2022 chage key-mapping and sleep related code.
*/

#include "xd87.h"
#include "hid_keycode.h"
#include "hobo_nicola.h"
#include "hobo_sleep.h"

static const uint8_t FN_XD_START_DFU	= FN_EXTRA_START;
static const uint8_t FN_RGB_BRIGHTER	= FN_EXTRA_START + 1;
static const uint8_t FN_RGB_DIMMER		= FN_EXTRA_START + 2;

// Function keys with Fn-key pressed.
static const uint16_t fn_keys[] PROGMEM = {
  HID_S     | WITH_R_CTRL , FN_SETUP_MODE,
	HID_U_ARROW, FN_MEDIA_VOL_UP,
	HID_D_ARROW, FN_MEDIA_VOL_DOWN,
	HID_R_ARROW, FN_MEDIA_SCAN_NEXT,
	HID_L_ARROW, FN_MEDIA_SCAN_PREV,
	HID_ENTER,	 FN_MEDIA_PLAY_PAUSE,
	HID_HOME,		 FN_MEDIA_PLAY_SKIP,
	HID_DELETE,	 FN_MEDIA_STOP,
	HID_END,		 FN_MEDIA_MUTE,
	HID_ESCAPE | WITH_R_CTRL,	 FN_SYSTEM_SLEEP,		// Ctrl + Fn + Esc 
	HID_ESCAPE | WITH_R_SHIFT | WITH_R_CTRL,	 FN_SYSTEM_POWER_DOWN,

	WITH_L_CTRL | WITH_R_CTRL | HID_B, FN_XD_START_DFU,
	WITH_R_CTRL | HID_D_ARROW, FN_RGB_DIMMER,
	WITH_R_CTRL | HID_U_ARROW, FN_RGB_BRIGHTER,
	
	HID_F1	, HID_IME_ON,
	HID_F2	, HID_IME_OFF,
	HID_F3	, HID_F14,
	HID_F4	, HID_F15,
	0, 0
};

class xd87Keyboard : public HoboNicola
{
  uint8_t _nicola_led;
public:
  xd87Keyboard() { 
		setting_value = Settings().get_extra();
		rgb_val = setting_value & 0xff;
   _nicola_led = 0;
	}

	// 兼用FnキーはAppキー
	virtual uint8_t get_Fn_key() const { return HID_APP; }

	virtual const uint16_t* get_fn_keys() { return fn_keys; }
	virtual void extra_function(uint8_t fk, bool pressed)	;

  void capslock_led(uint8_t on) { CAPSLED(on); }
  void scrlock_led(uint8_t on)  { BGLED(on); }
  void error_led(uint8_t on) { }
  void nicola_led(uint8_t on) {
    _nicola_led = on;
    scrlock_led(on);      
  }
  void toggle_nicola_led() {
    _nicola_led ^= 1;
    nicola_led(_nicola_led);
  }
	void rgb_on() ;
	void rgb_off() { set_rgb_value(0); }

private:
  int16_t rgb_val;
  static const uint8_t VAL_MAX = 0x7f; // 明るい
	uint32_t setting_value;	// とりあえず下位8bitでRGB LEDの明るさを。
  void rgb_on( uint8_t val);
  void rgb_dimmer();
  void rgb_brighter();
	void save_rgb();
};

xd87Keyboard xd87;

void xd87Keyboard::save_rgb() {
	rgb_val &= 0xff;
	setting_value = (setting_value & 0xffffff00) | rgb_val;
	Settings().save_extra(setting_value);
}

void xd87Keyboard::rgb_on(uint8_t val) {
  rgb_val = val;
	rgb_on();
}

void xd87Keyboard::rgb_on( ) {
	set_rgb_value(rgb_val);
	save_rgb();
}

#define rgb_unit 4
void xd87Keyboard::rgb_dimmer() {
  rgb_val -= rgb_unit;
  if (rgb_val < 0)
    rgb_val = 0;
  rgb_on();
}

void xd87Keyboard::rgb_brighter() {
  rgb_val += rgb_unit;
  if (rgb_val > VAL_MAX)
    rgb_val = VAL_MAX;
  rgb_on();
}

void xd87Keyboard::extra_function(uint8_t k, bool pressed) {
  if (!pressed)
    return;  
  switch(k) {
	case FN_XD_START_DFU:
		releaseAll();	
		delay(10);
		enter_dfu_bootloader();	// never return.
		break;
  case FN_RGB_DIMMER:
    rgb_dimmer();  
    break;
  case FN_RGB_BRIGHTER:
    rgb_brighter(); 
    break;
  default:
    break;
  }
}

void setup() {
	init_xd87();
  BGLED(0);
  CAPSLED(0);
  xd87Keyboard::init_hobo_nicola(&xd87);
}

void loop() {
  bool pressed;
  uint8_t key = xd87_get_key(pressed, Settings().is_us_layout());
  if (key != 0 && !is_usb_suspended()) {
    xd87.key_event(key, pressed);
		xd87.rgb_on();
		return;
	}
	xd87.idle();
	
	if (is_usb_suspended() ) {
		if (key != 0) {
			usb_wakeup();
			return;
		}
// start power save seq.
		xd87.capslock_led(false);
		xd87.nicola_led(false);
		xd87.rgb_off();

		DDRB = DDRC = DDRD = DDRE = DDRF = 0; 					// all input
		PORTB = PORTC = PORTD = PORTE = PORTF = 0xff;	// all input-pullup
		enter_sleep(500);	// 
		init_xd87();
		matrix_scan();	// キースキャンさせる
	} else
		enter_sleep();	 // TC0のオーバーフロー割込み(2msec周期)で目覚める。		
}
